/*
 * Generic visitor library
 *
 * Copyright (C) 1998-2001 Jose' Orlando Pereira, Universidade do Minho
 */
/* XTL2 - the eXternalization Template Library v. 2
 * http://xtl.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 *
 * $Id: visitors.h,v 1.5 2001/01/04 11:12:53 jop Exp $
 */


#ifndef __XTL2_VISITORS_H
#define __XTL2_VISITORS_H

namespace xtl {

// visitor traits

class destructor {};
class constructor {};

// char traits

class ascii {};
class raw {};

// adaptors

template <class Type>
struct _reference_adaptor {
	typedef Type type;

	Type*& ptr;

	_reference_adaptor(Type*& p):ptr(p) {}

	template <class Visitor>
	void composite(Visitor& visitor, destructor*) {
		visit(visitor, (typename Visitor::kind*)0, *ptr);
	}

	template <class Visitor>
	void composite(Visitor& visitor, constructor*) {
		ptr=new Type;
		visitor>>*ptr;
	}
};

template <class Type>
inline _reference_adaptor<Type> reference(Type*& p) {
	return _reference_adaptor<Type>(p);
}

template <class Type>
struct _pointer_adaptor: public _reference_adaptor<Type> {
	_pointer_adaptor(Type*& p):_reference_adaptor<Type>(p) {}

	template <class Visitor>
	void composite(Visitor& visitor, destructor*) {
		bool flag=ptr!=NULL;
		visitor.visit(flag);
		if (flag)
			_reference_adaptor<Type>::composite(visitor,(typename Visitor::kind*)0);
	}

	template <class Visitor>
	void composite(Visitor& visitor, constructor*) {
		bool flag;
		visitor.visit(flag);
		if (flag)
			_reference_adaptor<Type>::composite(visitor,(typename Visitor::kind*)0);
		else
			ptr=NULL;
	}
};

template <class Type>
inline _pointer_adaptor<Type> pointer(Type*& p) {
	return _pointer_adaptor<Type>(p);
}

struct _var_cstring_adaptor {
	char*& ptr;

	_var_cstring_adaptor(char*& p):ptr(p) {}

	template <class Visitor>
	void composite(Visitor& visitor, destructor*) {
		int size=std::strlen(ptr);
		visitor.visit(size);
		visitor.visit(ptr, size,(ascii*)0);
	}

	template <class Visitor>
	void composite(Visitor& visitor, constructor*) {
		int size;
		visitor.visit(size);
		ptr=new char[size+1];
		visitor.visit(ptr, size,(ascii*)0);
		ptr[size]=0;
	}
};

inline _var_cstring_adaptor cstring(char*& p) {
	return _var_cstring_adaptor(p);
}

struct _fixed_cstring_adaptor {
	char* ptr;
	int size;

	_fixed_cstring_adaptor(char* p, int s):ptr(p),size(s) {}

	template <class Visitor>
	void composite(Visitor& visitor, destructor*) {
		int effsize=std::strlen(ptr);
		visitor.visit(effsize);
		visitor.visit(ptr, effsize, (ascii*)0);
	}

	template <class Visitor>
	void composite(Visitor& visitor, constructor*) {
		int effsize;
		visitor.visit(effsize);
		if (effsize>size)
			; // do something
		visitor.visit(ptr, effsize, (ascii*)0);
		ptr[effsize]=0;
	}
};

inline _fixed_cstring_adaptor cstring(char* p, int size) {
	return _fixed_cstring_adaptor(p, size);
}

template <class Type, class Idx>
struct _vector_adaptor {
	Type* ptr;
	Idx size;

	_vector_adaptor(Type* p, Idx s):ptr(p),size(s) {}

	template <class Visitor, class Kind>
	void composite(Visitor& visitor, Kind kind) {
		for(Idx i=0;i<size;i++)
			visitor>>ptr[i];
	}
};

template <class Type, class Idx>
inline _vector_adaptor<Type,Idx> vector(Type* p, Idx size) {
	return _vector_adaptor<Type,Idx>(p, size);
}

template <class Type, class Idx>
struct _array_adaptor {
	Type*& ptr;
	Idx& size;

	_array_adaptor(Type*& p, Idx& s):ptr(p),size(s) {}

	template <class Visitor>
	void composite(Visitor& visitor, destructor*) {
		visitor.visit(size);
		visitor>>vector(ptr,size);
	}

	template <class Visitor>
	void composite(Visitor& visitor, constructor*) {
		visitor.visit(size);
		ptr=new Type[size];
		visitor>>vector(ptr,size);
	}
};

template <class Type, class Idx>
inline _array_adaptor<Type,Idx> array(Type*& p, Idx& size) {
	return _array_adaptor<Type,Idx>(p, size);
}

template <class Type, class Idx>
struct _opaque_adaptor {
	Type* ptr;
	Idx size;

	_opaque_adaptor(Type* p, Idx s):ptr(p),size(s) {}

	template <class Visitor, class Kind>
	void composite(Visitor& visitor, Kind) {
		visitor.visit(static_cast<char*>(ptr),size,(raw*)0);
	}
};

template <class Type, class Idx>
inline _opaque_adaptor<Type,Idx> opaque(Type* p, Idx size) {
	return _opaque_adaptor<Type,Idx>(p, size);
}

template <class Type, class Idx>
struct _bytes_adaptor {
	Type*& ptr;
	Idx& size;

	_bytes_adaptor(Type*& p, Idx& s):ptr(p),size(s) {}

	template <class Visitor>
	void composite(Visitor& visitor, destructor*) {
		visitor.visit(size);
		visitor.visit(static_cast<char*>(ptr),size,(raw*)0);
	}

	template <class Visitor>
	void composite(Visitor& visitor, constructor*) {
		visitor.visit(size);
		ptr=new Type[size];
		visitor.visit(static_cast<char*>(ptr),size,(raw*)0);
	}
};

template <class Type, class Idx>
inline _bytes_adaptor<Type,Idx> bytes(Type*& p, Idx& size) {
	return _bytes_adaptor<Type,Idx>(p, size);
}

};

// top level utilities

template <class Visitor, class Type>
inline Visitor& operator>>(Visitor& visitor, Type const& data) {
	visit(visitor, (typename Visitor::kind*)0, const_cast<Type&>(data));
	return visitor;
}

template <class Visitor, class Kind, class Type>
inline void visit(Visitor& visitor, Kind kind, Type& data) {
	data.composite(visitor, kind);
}

#define _def_visit_simple(type) \
template <class Visitor, class Kind>\
inline void visit(Visitor& visitor, Kind kind, type& data) {\
	visitor.visit(data);\
}

_def_visit_simple(bool)
_def_visit_simple(char)
_def_visit_simple(unsigned char)
_def_visit_simple(int)
_def_visit_simple(unsigned int)
_def_visit_simple(short)
_def_visit_simple(unsigned short)
_def_visit_simple(long)
_def_visit_simple(unsigned long)
_def_visit_simple(longlong)
_def_visit_simple(unsignedlonglong)
_def_visit_simple(float)
_def_visit_simple(double)

#undef _def_visit_simple

#endif
